;; Wb-tree File Based Associative String Data Base System.
;; Copyright (C) 1991, 1992, 1993, 2000, 2003 Free Software Foundation, Inc.
;;
;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this software; see the file COPYING.  If not, write to
;; the Free Software Foundation, 59 Temple Place, Suite 330, Boston,
;; MA 02111, USA.
;;
;; As a special exception, the Free Software Foundation gives permission
;; for additional uses of the text contained in its release of SCM.
;;
;; The exception is that, if you link the SCM library with other files
;; to produce an executable, this does not by itself cause the
;; resulting executable to be covered by the GNU General Public License.
;; Your use of that executable is in no way restricted on account of
;; linking the SCM library code into it.
;;
;; This exception does not however invalidate any other reasons why
;; the executable file might be covered by the GNU General Public License.
;;
;; This exception applies only to the code released by the
;; Free Software Foundation under the name SCM.  If you copy
;; code from other Free Software Foundation releases into a copy of
;; SCM, as the General Public License permits, the exception does
;; not apply to the code that you add in this way.  To avoid misleading
;; anyone as to the status of such modified files, you must delete
;; this exception notice from them.
;;
;; If you write modifications of your own for SCM, it is your choice
;; whether to permit this exception to apply to your modifications.
;; If you do not wish that, delete this exception notice.

;;; routines in this file return pass/fail (#t/#f)

(require 'i/o-extensions)
(load (in-vicinity (program-vicinity) "wbsys.scm"))

(pragma.c "#ifdef unix"
	  "# include <sys/types.h>"
	  "# include <sys/stat.h>"
	  "# include <fcntl.h>"
	  "# include <unistd.h>"
	  "#endif")

(define io-diag #f)

(define (init-filesystem!) #f)

(define (min-file-blk-size name) 128)

(define (blk:file-create name bsiz)
  (open-output-file name))

(define (blk:file-open-modify name bsiz)
  (open-io-file name))

(define (blk:file-open-read-only name bsiz)
  (open-input-file name))

(define (blk:file-close file) (close-port file))

(define (blk:read file blk bsiz blknum)
  (file-set-position file (* bsiz blknum))
  (cond ((= bsiz (substring-read! blk 0 bsiz file))
	 (if io-diag (dprintf "rd:%*s%c %10ld\n"
			      (+ 1 (- (BLK:LEVEL blk) LEAF)) ""
			      (BLK:TYP blk) blknum))
	 (if (BLK:TYP? blk FRL-TYP)
	     (set! read-fl-ct (+ read-fl-ct 1))
	     (set! read-ct (+ read-ct 1)))
	 #t)
	(else (edprintf " couldn't read blk %ld\n" blknum)
	      #f)))

(define (blk:write file blk bsiz blknum)
  (file-set-position file (* bsiz blknum))
  (cond ((= bsiz (substring-write blk 0 bsiz file))
	 (if io-diag (dprintf "wr:%*s%c %10ld\n"
			      (+ 1 (- (BLK:LEVEL blk) LEAF)) ""
			      (BLK:TYP blk) blknum))
	 (if (BLK:TYP? blk FRL-TYP)
	     (set! write-fl-ct (+ write-fl-ct 1))
	     (set! write-ct (+ write-ct 1)))
	 #t)
	(else (edprintf " couldn't write blk %ld\n" blknum)
	      #f)))

(define (extend-file file blk bsiz blknum)
  (file-set-position file (* bsiz blknum))
  (cond ((= bsiz (substring-write blk 0 bsiz file))
	 (if io-diag (dprintf "Extending file %d blks\n"
			      (+ (quotient FLC-LEN 2) 1)))
	 #t)
	(else (edprintf " couldn't extend file %ld\n" blknum)
	      #f)))
