;; WB-tree File Based Associative String Data Base System.
;; Copyright (C) 1991, 1992, 1993, 2000 Free Software Foundation, Inc.
;;
;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this software; see the file COPYING.  If not, write to
;; the Free Software Foundation, 59 Temple Place, Suite 330, Boston,
;; MA 02111, USA.
;;
;; As a special exception, the Free Software Foundation gives permission
;; for additional uses of the text contained in its release of SCM.
;;
;; The exception is that, if you link the SCM library with other files
;; to produce an executable, this does not by itself cause the
;; resulting executable to be covered by the GNU General Public License.
;; Your use of that executable is in no way restricted on account of
;; linking the SCM library code into it.
;;
;; This exception does not however invalidate any other reasons why
;; the executable file might be covered by the GNU General Public License.
;;
;; This exception applies only to the code released by the
;; Free Software Foundation under the name SCM.  If you copy
;; code from other Free Software Foundation releases into a copy of
;; SCM, as the General Public License permits, the exception does
;; not apply to the code that you add in this way.  To avoid misleading
;; anyone as to the status of such modified files, you must delete
;; this exception notice from them.
;;
;; If you write modifications of your own for SCM, it is your choice
;; whether to permit this exception to apply to your modifications.
;; If you do not wish that, delete this exception notice.

(load (in-vicinity (program-vicinity) "wbsys.scm"))

;;; PERFORMANCE STATISTICS

(define next-ct #f)
(define next-fct #f)
(define prev-ct #f)
(define prev-fct #f)
(define get-ct #f)
(define get-fct #f)
(define put-ct #f)
(define put-fct #f)
(define rem-ct #f)
(define rem-fct #f)
(define ge-ct #f)
(define ge-fct #f)
(define tge-ct #f)
(define tge-fct #f)
(define tce-ct #f)
(define tce-fct #f)

(define chains-to-next #f)		; counters for special-case occurrences
(define deferred-inserts #f)
(define split-index-inserts #f)
(define index-screw-case  #f)
(define block-splits #f)
(define block-deletes #f)
(define deferred-deletes #f)
(define dir-dty-ct #f)

(define read-ct #f)			; buffer i/o counters
(define write-ct #f)
(define read-fl-ct #f)
(define write-fl-ct #f)
(define flush-ct #f)

(define (clear-stats)
  (set! next-ct 0)
  (set! next-fct 0)
  (set! prev-ct 0)
  (set! prev-fct 0)
  (set! get-ct 0)
  (set! get-fct 0)
  (set! put-ct 0)
  (set! put-fct 0)
  (set! rem-ct 0)
  (set! rem-fct 0)
  (set! ge-ct 0)
  (set! ge-fct 0)
  (set! tge-ct 0)
  (set! tge-fct 0)
  (set! tce-ct 0)
  (set! tce-fct 0)

  (set! chains-to-next 0)
  (set! deferred-inserts 0)
  (set! split-index-inserts 0)
  (set! index-screw-case  0)
  (set! block-splits 0)
  (set! block-deletes 0)
  (set! deferred-deletes 0)
  (set! dir-dty-ct 0)

  (set! read-ct 0)
  (set! write-ct 0)
  (set! read-fl-ct 0)
  (set! write-fl-ct 0)
  (set! flush-ct 0)
  SUCCESS)

(define (cstats)
  (stats)
  (clear-stats))

(define (stats)
  (dprintf "\n" )
  (dprintf "OPERATIONS SUMMARY:\n" )
  (dprintf
	   "            NEXT     PREV      GET      PUT      REM   GETENT   CACHE    CACHE\n")
  (dprintf
	   "                                                                AVAIL?   VALID?\n")
  (dprintf "   succ %8lu %8lu %8lu %8lu %8lu %8lu %8lu %8lu\n"
	   next-ct prev-ct get-ct put-ct rem-ct ge-ct tge-ct tce-ct)
  (dprintf "   fail %8lu %8lu %8lu %8lu %8lu %8lu %8lu %8lu\n\n"
	   next-fct prev-fct get-fct put-fct rem-fct ge-fct tge-fct tce-fct)

  (dprintf "INTERNAL OPERATIONS SUMMARY:\n" )
  (dprintf "   chains-to-next = %d\n" chains-to-next)
  (dprintf "   %d block splits, %d deferred parent updates, %d undone\n"
	   block-splits deferred-inserts  deferred-inserts)
  (dprintf "   %d split index inserts; %d index-insert screw cases \n"
	   split-index-inserts index-screw-case )
  (dprintf "   %d block deletes; %d deferred block deletes\n\n"
	   block-deletes deferred-deletes)

  (dprintf "I/O SUMMARY: %lu READS, %lu WRITES, %lu FLUSH (of dirty bufs) calls.\n"
	   read-ct write-ct flush-ct)
  (dprintf "             %lu FREELIST READS, %lu FREELIST WRITES.\n"
	   read-fl-ct write-fl-ct)
  (dprintf "             %lu DIRS LEFT DTY.\n"
	   dir-dty-ct)
  (let ((ops (max (+ next-fct prev-fct get-fct put-fct rem-fct
		 next-ct prev-ct get-ct put-ct rem-ct ) 1)))
    (dprintf "   AVG BLKS REFERENCED PER OPERATION: %7lu%%\n"
	   (quotient (* 100 (+ ge-ct tce-ct)) ops))
    (dprintf "   AVG DISK I/Os PER OPERATION: %7lu%%; "
	     (quotient (* 100 (+ read-ct write-ct read-fl-ct write-fl-ct)) ops))
    (dprintf "   WRITE/READ RATIO: %7lu%%\n"
	     (quotient (* 100 (+ write-ct write-fl-ct)) (max 1 (+  read-ct read-fl-ct))))
    (if (> put-ct 0)
	(dprintf "   AVG DISK WRITES PER PUT: %7lu%%\n"
		 (quotient (* 100 write-ct) put-ct)))
    )
  (dprintf "\n" )
  (dprintf "MODES IN EFFECT:\n" )
  (dprintf "%lu buffers [hashed over %d buckets]; blksize (max)=%d; FLC-LEN=%d\n"
	   num-ents-ct NUM-BUKS blk-size FLC-LEN)
  (dprintf "defer-insert-updates= %2s; defer-block-deletes=%2s\n"
	   (if defer-insert-updates "#t" "#f") (if defer-block-deletes "#t" "#f") )
  (dprintf "cache-ent-enable=%2s; clever-cache-enable=%2s\n"
	   (if cache-ent-enable "#t" "#f") (if clever-cache-enable "#t" "#f") )
  (dprintf "\n" )
  SUCCESS)

(define (show-buffer! ent)
  (dprintf "SEG = %d ID= %lu ACC= %d REF= %d DTY= %d AGE= %d\n"
	   (ENT:SEG ent)(ENT:ID ent)(ENT:ACC ent)(ENT:REF ent)(ENT:DTY? ent)(ENT:AGE ent)))

(define buf-verbose #t)

(define (show-buffer-1 ent)
  (cond ((or buf-verbose (> (ENT:SEG ent) -1))
	 (dprintf " %3d:%-6lu %6lu %8d %6u %7d %3d %4d"
		  (ENT:SEG ent)
		  (ENT:ID ent)
		  (HASH2INT (ENT:SEG ent) (ENT:ID ent))
		  (ENT:ACC ent)
		  (ENT:REF ent)
		  0
		  (ENT:DTY? ent)
		  (ENT:AGE ent))
	 (if (> (ENT:SEG ent) -1)
	     (dprintf " %5d %4c\n"
		      (- (BLK:LEVEL (ENT:BLK ent)) LEAF)
		      (BLK:TYP (ENT:BLK ent)))
	     (dprintf "\n"))
	 ))
  SUCCESS)

(define (show-buffers)
  (dprintf "\n" )
  (dprintf
	   " SEG:ID        BUK      ACC    REF READERS DTY  AGE LEVEL TYPE\n")
  (do-seg-buffers -1 show-buffer-1))

(define (sb) (show-buffers))
